@@ -1,6 +1,8 @@
+from __future__ import annotations
+
 from dateutil.parser import isoparse
 
-from hcloud.core.domain import BaseDomain
+from ..core import BaseDomain
 
 
 class DeprecationInfo(BaseDomain):
@@ -22,8 +24,8 @@ class DeprecationInfo(BaseDomain):
 
     def __init__(
         self,
-        announced=None,
-        unavailable_after=None,
+        announced: str | None = None,
+        unavailable_after: str | None = None,
     ):
         self.announced = isoparse(announced) if announced else None
         self.unavailable_after = (
